import cython
cimport numpy as np
import numpy as np

ctypedef unsigned long long ULLong
ctypedef unsigned int UInt


cpdef ULLong row_idx_from_mat_coords(ULLong n, ULLong i, ULLong j):
    if i <= j:
        return <ULLong> (i * n - (i - 1) * i / 2 + j - i)
    else:
        return <ULLong> (j * n - (j - 1) * j / 2 + i - j)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef UInt n_cartesian(UInt[:] groupings):
    cdef UInt total, i, i_start, i_end, i_size, j, n_groups

    n_groups = groupings.shape[0] - 1
    if n_groups == 1:
        return groupings[1]

    total = 0
    for i in range(n_groups):
        i_start = groupings[i]
        i_end = groupings[i + 1]
        i_size = i_end - i_start
        for j in range(i + 1, n_groups):
            total += i_size * (groupings[j+1] - groupings[j])
    return total


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef cartesian_sum(ULLong n_indices, UInt[:] groupings, UInt[:] group_idxs, double[:] group_vals, double[:] data):

    cdef UInt n_comb = n_cartesian(groupings)
    cdef np.ndarray[np.uint64_t, ndim=1] idx_arr = np.empty(n_comb, dtype=np.uint64)
    cdef np.ndarray[np.float64_t, ndim=1] idx_val = np.empty(n_comb, dtype=np.float64)

    cdef UInt i, i_start_idx, i_end_idx, j, j_start_idx, j_end_idx, cur_i, tax_i, cur_j, n_iter, n_groups
    cdef double dist_i, dist_j

    n_iter = 0
    for i in range(groupings.shape[0] - 1):
        i_start_idx = groupings[i]
        i_end_idx = groupings[i+1]

        for j in range(i):
            j_start_idx = groupings[j]
            j_end_idx = groupings[j + 1]

            for cur_i in range(i_start_idx, i_end_idx):
                tax_i = group_idxs[cur_i]
                dist_i = group_vals[cur_i]
                for cur_j in range(j_start_idx, j_end_idx):
                    data[row_idx_from_mat_coords(n_indices, tax_i, group_idxs[cur_j])] = dist_i + group_vals[cur_j]
                    n_iter += 1
    return n_iter



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from collections import deque, defaultdict
from typing import Tuple, Union

import dendropy
import h5py
import numpy as np
from tqdm import tqdm

from phylodm.common import create_mat_vector, compact_int_mat
from phylodm.indices import Indices
from phylodm.pdm_c import cartesian_sum
from phylodm.symmat import SymMat


class PDM(SymMat):

    def __init__(self):
        """Variables are set when called by any of the static methods."""
        super().__init__()
        self._tree_length = None
        self._method = None

    @staticmethod
    def get_from_dendropy(tree: dendropy.Tree, method: str = 'pd') -> 'PDM':
        """Create a PDM from a dendropy tree object."""
        return PDM()._get_from_dendropy(tree, method)

    @staticmethod
    def get_from_newick_file(path_tree: str, method: str = 'pd') -> 'PDM':
        """Create a PDM from a newick file."""
        tree = dendropy.Tree.get_from_path(path_tree,
                                           schema='newick',
                                           rooting='force-unrooted',
                                           preserve_underscores=True)
        return PDM.get_from_dendropy(tree, method)

    @staticmethod
    def get_from_path(path: str) -> 'PDM':
        """Load a PDM from a cache."""
        return PDM()._get_from_path(path)

    @staticmethod
    def _get_int_node_depth(tree):
        out = defaultdict(set)
        queue = deque([(0, tree.seed_node)])
        while len(queue) > 0:
            depth, node = queue.pop()
            out[depth].add(node)
            for child_node in node.child_node_iter():
                if not child_node.is_leaf():
                    queue.append((depth + 1, child_node))
        return out

    @staticmethod
    def _preprocess_tree(tree: dendropy.Tree) -> dendropy.Tree:
        """Deep copy and pre-process a DendroPy tree."""
        out = dendropy.Tree.get_from_string(tree.as_string(schema='newick'), 'newick')
        if tree.seed_node.parent_node is not None:
            raise Exception('DendroPy did not seed the tree correctly.')
        return out

    def _get_from_dendropy(self, tree: dendropy.Tree, method: str = 'pd') -> 'PDM':

        # Initialise the tree
        tree = PDM._preprocess_tree(tree)

        # Initialise the class based on arguments.
        self._method = method
        if method == 'pd':
            use_pd = True
            self._d_type = np.dtype('float64')
            self._arr_default = 0.0
            self._tree_length = tree.length()
        elif method == 'node':
            use_pd = False
            self._d_type = np.dtype('uint32')
            self._arr_default = 0
            self._tree_length = len(tree.edges())
        else:
            raise NotImplemented(f'Unknown method: {method}')

        # Get each node at a specific depth.
        depth_to_node = PDM._get_int_node_depth(tree)

        # Pre-process each leaf node and create the indices.
        self._indices = Indices()
        for leaf_node in sorted(tree.leaf_node_iter(), key=lambda x: x.taxon.label):
            leaf_idx = self._indices.add_key(leaf_node.taxon.label)
            leaf_node.attr_child_dist = [(leaf_idx, leaf_node.edge_length if use_pd else 1)]

        # Create the vector to store the results.
        self._data = create_mat_vector(len(self._indices), 0.0)

        # Process the deepest nodes first, merging data at each level.
        with tqdm(total=len(self._data), unit_scale=True) as p_bar:
            n_indices = len(self._indices)
            for cur_depth, cur_nodes in sorted(depth_to_node.items(), key=lambda x: -x[0]):
                for node in cur_nodes:

                    # Extract the groups
                    groupings = [0]
                    group_idxs = list()
                    group_vals = list()
                    child_dist, child_groups = list(), list()
                    offset = 0

                    child_groups = list()

                    for child_node in node.child_node_iter():
                        child_groups.append(child_node.attr_child_dist)
                        for desc_idx, desc_dist in child_node.attr_child_dist:
                            group_idxs.append(desc_idx)
                            group_vals.append(desc_dist)
                            offset += 1
                        groupings.append(offset)
                        child_dist.extend(child_node.attr_child_dist)
                        del child_node.attr_child_dist

                    groupings = np.array(groupings, dtype=np.uint32)
                    group_idxs = np.array(group_idxs, dtype=np.uint32)
                    group_vals = np.array(group_vals, dtype=np.float64)

                    n_iter = cartesian_sum(n_indices, groupings, group_idxs, group_vals, self._data)
                    p_bar.update(n_iter)

                    # Record the distance from this node to its leaf nodes, and bring up.
                    # Add th distance from thsi node to its children
                    if cur_depth > 0:
                        node.attr_child_dist = [(x, y + (node.edge_length if use_pd else 1)) for x, y in child_dist]

        # Use the smallest possible data type for the matrix
        if method == 'node':
            self._data = compact_int_mat(self._data)
            self._d_type = self._data.dtype
        return self

    def as_matrix(self, normalised: bool = False) -> Tuple[Tuple[str], np.array]:
        """Return the PDM as a symmetric numpy matrix."""
        labels, mat = super().as_matrix()
        if normalised:
            mat = mat * (1.0 / self._tree_length)
        return labels, mat

    def save_to_path(self, path: str):
        """Write the PDM to a cache."""
        with h5py.File(path, 'w') as f:
            f.create_dataset('indices',
                             data=[t.encode('ascii') for t in self._indices.get_keys()],
                             dtype=h5py.string_dtype(encoding='ascii'))
            f.create_dataset('data', data=self._data, dtype=self._d_type)
            f.create_dataset('tree_length', data=self._tree_length)
            f.create_dataset('arr_default', data=self._arr_default)
            f.create_dataset('method', data=self._method.encode('ascii'), dtype=h5py.string_dtype(encoding='ascii'))

    def _get_from_path(self, path: str) -> 'PDM':
        """Read the PDM from a cache."""
        self._indices = Indices()
        with h5py.File(path, 'r') as hf:
            self._arr_default = hf['arr_default'][()]
            self._data = hf['data'][()]
            self._tree_length = hf['tree_length'][()]
            self._method = hf['method'][()].decode('utf-8')
            for idx, key in enumerate(hf['indices'][()]):
                self._indices.add_key(key.decode('utf-8'))
        self._d_type = self._data.dtype
        return self

    def get_value(self, key_i: str, key_j: str, normalised: bool = False) -> Union[int, float]:
        """Get the value of two keys."""
        value = super().get_value(key_i, key_j)
        if normalised:
            value *= 1 / self._tree_length
        return value



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import argparse
import os
import sys

from phylodm import __version__
from phylodm.pipeline import newick_to_pdm


def print_help():
    lines = [f'phylodm v{__version__}',
             'usage: [path to newick] [pd/node] [path to output]']
    print('\n'.join(lines))


def main(args=None):
    parser = argparse.ArgumentParser(description='Compute a phylogenetic '
                                                 'distance matrix for a given '
                                                 'newick tree.')
    parser.add_argument('newick', type=str, help='path to the newick file')
    parser.add_argument('method', type=str, help='use patristic distance {pd} or node distance {node}')
    parser.add_argument('output', type=str, help='path to output matrix')

    # Verify that a subparser was selected
    if len(sys.argv) == 1:
        print_help()
        sys.exit(0)
    elif sys.argv[1] in {'-v', '--v', '-version', '--version'}:
        print(f'phylodm v{__version__}')
        sys.exit(0)
    else:
        print(f'phylodm v{__version__}')
        args = parser.parse_args(args)

        # Validate the input arguments.
        if not os.path.isfile(args.newick):
            print(f'The input file cannot be found: {args.newick}')
            sys.exit(1)
        if args.method not in {'pd', 'node'}:
            print('Invalid choice for method, select either: pd or node')
            sys.exit(1)
        out_dir = os.path.dirname(args.output)
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)

        newick_to_pdm(args.newick, args.method, args.output)

    # Done - no errors.
    sys.exit(0)


if __name__ == "__main__":
    main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from typing import Union

import numpy as np


def create_mat_vector(n: int, default: Union[int, float]) -> np.array:
    """Create a vector which represents the upper/lower triangle of a matrix."""
    return np.full(n * (n + 1) // 2, default)


def compact_int_mat(mat: np.array) -> np.array:
    """For a matrix of integers, return a copy with the smallest data type."""
    signed = (('int8', -128, 127),
              ('int16', -32768, 32767),
              ('int32', -2147483648, 2147483647),
              ('int64', -9223372036854775808, 9223372036854775807))
    unsigned = (('uint8', 0, 255),
                ('uint16', 0, 65535),
                ('uint32', 0, 4294967295),
                ('uint64', 0, 18446744073709551615))
    m_min, m_max = mat.min(), mat.max()
    if m_min < 0:
        options = signed
    else:
        options = unsigned
    for d_type, d_min, d_max in options:
        if m_min >= d_min and m_max <= d_max:
            return mat.astype(np.dtype(d_type))
    return mat



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from typing import Collection, Tuple, Optional, Union

import h5py
import numpy as np
from phylodm.pdm_c import row_idx_from_mat_coords

from phylodm.common import create_mat_vector
from phylodm.indices import Indices


class SymMat(object):

    def __init__(self):
        """Variables are set when called by any of the static methods."""
        self._d_type: Optional[np.dtype] = None
        self._arr_default: Optional[Union[int, float]] = None
        self._indices: Optional[Indices] = None
        self._data: Optional[np.array] = None

    def __eq__(self, other) -> bool:
        """Two SymMats are equal if the data, defaults, and indices are equal."""
        if isinstance(other, SymMat):
            return self._d_type == other._d_type and \
                   self._arr_default == other._arr_default and \
                   self._indices == other._indices and \
                   np.array_equal(self._data, other._data)
        return False

    @staticmethod
    def get_from_shape(n_indices: int, d_type: np.dtype,
                       arr_default: Union[int, float] = 0) -> 'SymMat':
        """Create a blank SymMat given the specifications."""
        return SymMat()._get_from_shape(n_indices, d_type, arr_default)

    @staticmethod
    def get_from_indices(indices: Collection[str], d_type: np.dtype,
                         arr_default: Union[int, float] = 0) -> 'SymMat':
        """Create a blank SymMat given the indices."""
        return SymMat()._get_from_indices(indices, d_type, arr_default)

    @staticmethod
    def get_from_path(path: str) -> 'SymMat':
        """Load the SymMat from a cache."""
        return SymMat()._get_from_path(path)

    def _idx_from_key(self, key_i: str, key_j: str) -> int:
        """Determine the row vector index given the indices names."""
        i = self._indices.get_key_idx(key_i)
        j = self._indices.get_key_idx(key_j)
        return row_idx_from_mat_coords(len(self._indices), i, j)

    def _get_from_shape(self, n_indices: int, d_type: np.dtype,
                        arr_default: Union[int, float] = 0) -> 'SymMat':
        """Create a blank SymMat given the specifications."""
        self._d_type = d_type
        self._arr_default = arr_default
        self._indices = Indices()
        self._data = create_mat_vector(n_indices, arr_default)
        return self

    def _get_from_indices(self, indices: Collection[str], d_type: np.dtype,
                          arr_default: Union[int, float] = 0) -> 'SymMat':
        """Create a blank SymMat given the indices."""
        self._d_type = d_type
        self._arr_default = arr_default
        self._indices = Indices()
        self._indices.add_keys(indices)
        self._data = create_mat_vector(len(indices), arr_default)
        return self

    def _get_from_path(self, path: str) -> 'SymMat':
        """Load the SymMat from a cache."""
        self._indices = Indices()
        with h5py.File(path, 'r') as hf:
            self._arr_default = hf['arr_default'][()]
            self._data = hf['data'][()]
            for idx, key in enumerate(hf['indices'][()]):
                self._indices.add_key(key.decode('utf-8'))
        self._d_type = self._data.dtype
        return self

    def save_to_path(self, path: str):
        """Save the SymMat to a cache."""
        with h5py.File(path, 'w') as f:
            f.create_dataset('indices',
                             data=[t.encode('ascii') for t in self._indices.get_keys()],
                             dtype=h5py.string_dtype(encoding='ascii'))
            f.create_dataset('data', data=self._data, dtype=self._d_type)
            f.create_dataset('arr_default', data=self._arr_default)

    def get_value(self, key_i: str, key_j: str) -> Union[int, float]:
        """Get the value for the specified keys."""
        data_idx = self._idx_from_key(key_i, key_j)
        return self._data[data_idx]

    def set_value(self, key_i: str, key_j: str, value: Union[float, int]):
        """Set a specific value given the keys."""
        self._data[self._idx_from_key(key_i, key_j)] = value

    def remove_keys(self, keys: Collection[str]):
        """Remove the key and associated values from the matrix."""
        key_idx = self._indices.get_key_indices()
        keep_keys = sorted(key_idx.keys() - set(keys))
        new_indices = Indices()
        new_indices.add_keys(keep_keys)

        # Create the new matrix and import all of the keys across.
        new_mat = SymMat.get_from_indices(keep_keys, d_type=self._d_type,
                                          arr_default=self._arr_default)

        # Determine the mapping for importing the data across.
        for i in range(len(keep_keys)):
            for j in range(i + 1):
                new_mat.set_value(keep_keys[i], keep_keys[j],
                                  self.get_value(keep_keys[i], keep_keys[j]))
        self._data = new_mat._data
        self._indices = new_mat._indices

    def as_matrix(self) -> Tuple[Tuple[str], np.array]:
        """Return a symmetric numpy matrix given the SymMat."""
        n_indices = len(self._indices)
        mat = np.full((n_indices, n_indices), 0, dtype=self._d_type)
        mat[np.triu_indices_from(mat)] = self._data
        diag = mat[np.diag_indices_from(mat)]
        mat = mat + mat.T
        mat[np.diag_indices_from(mat)] = mat[np.diag_indices_from(mat)] - diag
        return self._indices.get_keys(), mat



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from typing import Dict, Collection, Tuple, List

from phylodm.exceptions import DuplicateIndex


class Indices(object):
    """Stores a unique list of strings, preserving their order and providing
    constant time lookup for the order in which they were added."""

    def __init__(self):
        """Instance variables are populated via the add key(s) method."""
        self._keys: List[str] = list()
        self._keys_idx: Dict[str, int] = dict()

    def __eq__(self, other) -> bool:
        """Equal if the indices and the positions are the same."""
        if isinstance(other, Indices):
            return self._keys == other._keys and \
                   self._keys_idx == other._keys_idx
        return False

    def __len__(self) -> int:
        """The number of indices stored."""
        return len(self._keys)

    def __contains__(self, key: str) -> bool:
        """Returns True if a key is in the list of indices."""
        return key in self._keys_idx

    def add_key(self, key: str):
        """Add a key, a DuplicateIndex is raised if the key is not unique."""
        if key in self._keys_idx:
            raise DuplicateIndex(f'Cannot add duplicate key: {key}')
        new_idx = len(self._keys)
        self._keys.append(key)
        self._keys_idx[key] = new_idx
        return new_idx

    def add_keys(self, keys: Collection[str]) -> List[int]:
        """Adds a collection of keys to the indices."""
        indexes = list()
        for key in keys:
            indexes.append(self.add_key(key))
        return indexes

    def get_keys(self) -> Tuple[str]:
        """Return a tuple of the indices stored."""
        return tuple(self._keys)

    def get_key_idx(self, key: str) -> int:
        """Returns the index of the specified key."""
        return self._keys_idx[key]

    def get_key_indices(self) -> Dict[str, int]:
        """Returns a dictionary of all keys and their indices."""
        return self._keys_idx



import os
import platform
import re
from distutils.command.build import build as build_orig

from setuptools import setup, Extension


def read_version():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'phylodm/__init__.py')
    with open(path) as fh:
        return re.search(r'__version__\s?=\s?[\'"](.+)[\'"]', fh.read()).group(1)


def readme():
    with open('README.md') as f:
        return f.read()


class build(build_orig):

    def finalize_options(self):
        super().finalize_options()
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        for extension in self.distribution.ext_modules:
            extension.include_dirs.append(numpy.get_include())
        from Cython.Build import cythonize
        self.distribution.ext_modules = cythonize(self.distribution.ext_modules,
                                                  language_level=3)


compile_extra_args = ['-O3']
link_extra_args = list()
# if platform.system() == "Windows":
#     compile_extra_args = ["/std:c++latest", "/EHsc"]
if platform.system() == "Darwin":
    compile_extra_args.extend(['-std=c++11', "-mmacosx-version-min=10.9"])
    link_extra_args.extend(["-stdlib=libc++", "-mmacosx-version-min=10.9"])

ext_modules = [Extension('phylodm.pdm_c', ['phylodm/pdm_c.pyx'],
                         language='c++',
                         extra_compile_args=compile_extra_args,
                         extra_link_args=link_extra_args
                         )]

setup(name='phylodm',
      version=read_version(),
      description='Efficient calculation of phylogenetic distance matrices.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      author='Aaron Mussig',
      author_email='aaronmussig@gmail.com',
      url='https://github.com/aaronmussig/PhyloDM',
      license='GPL3',
      project_urls={
          "Bug Tracker": "https://github.com/aaronmussig/PhyloDM/issues",
          "Documentation": "https://github.com/aaronmussig/PhyloDM",
          "Source Code": "https://github.com/aaronmussig/PhyloDM",
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='phylogenetic distance matrix symmetric',
      packages=['phylodm'],
      package_data={'phylodm': ['pdm_c.pyx']},
      entry_points={
          'console_scripts': [
              'phylodm = phylodm.__main__:main'
          ]
      },
      install_requires=['numpy', 'dendropy', 'h5py', 'tqdm'],
      setup_requires=['cython', 'numpy'],
      python_requires='>=3.6',
      data_files=[("", ["LICENSE"])],
      ext_modules=ext_modules,
      cmdclass={'build': build}
      )



from phylodm.pdm import PDM


def newick_to_pdm(path_newick: str, method: str, path_out: str):
    """Converts a newick file to a PDM and caches it to disk."""
    pdm = PDM.get_from_newick_file(path_newick, method)
    pdm.save_to_path(path_out)
